from rich import print as richprint
from owly_sdk.common.constants import *
from owly_sdk.common.request_config import RequestConfig
import requests
import time
import json
import yaml


def makerequest(_type,url,params=None,body=None,https_verify=False,header=None, config: RequestConfig = RequestConfig()):
    connection_try = config.connection_retry
    while True:
        try:
            if _type == 'post':
                r = requests.post(url, json=body, verify=https_verify)
            elif _type == 'get':
                r = requests.get(url, verify=https_verify,params=params,headers=header)
            elif _type == 'delete':
                r = requests.delete(url,verify=https_verify,params=params,headers=header)
            elif _type == 'put':
                if body is not None:
                    r = requests.put(url,verify=https_verify,json=body,headers=header)
                if params is not None:
                    r = requests.put(url,verify=https_verify,params=params,headers=header)
            status_code = r.status_code
            if status_code!=204:
                response = r.json()
            else:
                response = {}
            break
        except Exception as e:
            richprint(str(e))
            if connection_try>0:
                connection_try -=1
                time.sleep(config.connection_retry_time)
            else:
                response = {'status': 'error','reason':e}
                status_code = 500
                break
    return response, status_code


def format_response(response, output : str = None):
    """
    Format the response in the given format.
    
    :param response: The response to be formatted.
    :param output: The format of the output (json/yaml).
    
    :return: The formatted response
    """
    if output == 'yaml':
        return yaml.dump(response)
    if output == 'json':
        return json.dumps(response)
    return response


import typer
import yaml
import time
import uuid
import base64
import yaml
from pathlib import Path
from owly_sdk.common.utils import makerequest, format_response


class NsManager:
    """
    The NsManager class provides the functionality to manage Network Services (NS) such as creating, listing, describing, and deleting NS.
    """
    def __init__(self, domain_orch_url, request_config):
        """
        Initialize the NsManager class with the domain_orch_url and request_config.
        
        :param domain_orch_url: The URL of the Orchestrator.
        :param request_config: The RequestConfig object to be used for making requests.
        """
        self.domain_orch_url = domain_orch_url
        self.request_config = request_config
        
        
    def _nscreate(self, service,file):
        '''
        onboard network functions of a network service
        
        :param service: The network service description
        :param file: The path to the yaml file
        
        :return: The network service description
        '''
        if 'packageId' not in service.keys():
            service.update({'packageId':str(uuid.uuid4())})
        for vnf in service['vnfs']:
            for container in vnf['osContainerDesc']:
                if 'bootData' in container.keys():
                    for data in container['bootData']:
                        data.update({'data': (base64.b64encode((open(Path(file.parent) / data['file']).read()).encode('ascii'))).decode('ascii')})
        [ vnf.update({'vnfDId': str(uuid.uuid4())}) for vnf in service['vnfs'] ]
        if not isinstance(service['vimId'],list):
            service.update({'vimId':[service['vimId']]})
        r, s = makerequest(_type='post',url= f"{self.domain_orch_url}/create/ns",body=service)
        if s!=201:
            raise typer.Exit(code=1)
        return service


    def create_ns(self, file):
        """
        Creates a Network Service (NS) from a YAML file.
        
        :param file: The path to the YAML file containing the NS description.
        
        :return: None
        """
        
        try:
            # Convert to Path object if it's a string
            if isinstance(file, str):
                file = Path(file)

            with open(file) as f:
                services = yaml.safe_load_all(f)
                for service in services:
                    service = self._nscreate(service=service,file=file)
        except Exception as e:
            return e


    def _format_ns(self, list_ns, wide=False):
        """
        Formats the network service details for display.
        
        :param list_ns: The list of network services.
        :param wide: Indicates whether to display more details. Default is False.
        
        :return: The formatted network service details.
        """
        formatted_ns = []
        
        for ns in list_ns:
            vnfs = []
            duration = 0
        
            for vnf in ns['vnfs']:
                formatted_vnf = {
                    'vnfName': vnf['vnfName'],
                    'status': vnf['status'],
                }
                
                if vnf['status'] == 'INSTANTIATED':
                    formatted_vnf['interfaces'] = [
                        f"{networkDetail['ipaddress']}({networkDetail['interface']})"
                        for networkDetail in vnf['networkDetail']
                    ]
                    
                    onboarding_duration = float(vnf['onboardingDuration'])
                    instance_creation_duration = float(vnf['instanceCreationDuration'][:-1]) # Remove the 's' at the end
                    duration += onboarding_duration + instance_creation_duration

                    
                vnfs.append(formatted_vnf)

            ns = {
                'name': ns['name'],
                'status': ns['status'],
                'vnfs': vnfs
            }
            
            if wide:
                ns['duration'] = duration
                
            formatted_ns.append(ns)
        
        return formatted_ns


    def list_nss(self, name="all",vf_name="", latest="yes", wide=False):
        """
        Fetches the list of Network Functions (NF) from the API.
        
        :param name: The name of the NF to search for. Default is "all".
        :param vf_name: The name of the VF to search for. Default is "". NOT WORKING
        :param latest: Indicates whether only the latest NFs should be returned. Default is "yes". NOT WORKING
        :param wide: Indicates whether to display more details. Default is False.

        :return: The response from the API.
        """
        try:
            res, status_code = makerequest(
                _type='get',
                url=f'{self.domain_orch_url}/list/vnf',
                params={'name': name, 'vfName': vf_name, 'latest': latest}
            )
            
            if (isinstance(res, list) or isinstance(res, dict)) and status_code == 200:
                if 'status' in res and res['status'] == "NOTFOUND":
                    return res
                if name != "all":
                    res = [res]

                formatted_ns = self._format_ns(res, wide)
                if name == "all":
                    return formatted_ns
                return formatted_ns[0]
            
            return "Encountered an error in fetching the list of onboarded services"
        
        except Exception as e:
            return e


    def describe_ns(self, name: str, output: str = None):
        """
        Describe a network service NAME.

        :param name: The name of the network service to describe.
        :param output: Format of the output (json/yaml).
        
        :return: The response from the API or an error message.
        """
        try:
            response, status_code = makerequest(
                _type='get',
                url=f"{self.domain_orch_url}/list/vnf",
                params={'name': name}
            )

            if status_code == 200:
                return format_response(response, output)
            else:
                return Exception(response.get('message', 'Unknown error'))

        except Exception as e:
            return e


    def top(
        self,
        vim: str,
        name: str,
        vnf: str = None,
        output: str = None,
    ):
        """
        Performs top on network service VIM_NAME NAME.
        Provides information related to computational 
        resources consumed by the network service
        
        :param vim: The VIM ID
        :param name: The name of the network service
        :param vnf: The name of the VNF
        :param output: The format of the output (json/yaml). Default is 'json'.
        """
        r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/metrics/vnf",params={'name':name, 'vnfName':vnf, 'vimId':vim})
        if s in [200,404,400]:        
            return format_response(r, output)
        else:
            return "Encounted an error in fetching the list of onboarded services"
        

    def delete_ns(self, name: str):
        """
        Deletes a Network Function (NF) from the API.

        :param name: The name of the NF to delete.
        
        :return: None if the NF is deleted successfully, else an error message.
        """
        
        # Check if the NF exists
        r, s = makerequest(
                _type='get',
                url=f"{self.domain_orch_url}/list/vnf",
                params={'name':name,'latest':'no'}
            )
        
        if s == 200 and r['status'] == 'NOTFOUND':
            return r

        elif s == 200 and r['status'] in ['INSTANTIATED','TERMINATING']:
            package_id = r['packageId']
            vim_id = r['vimId']
            vnfs = [{'vnfName':vnf['vnfName'],'vnfDId':vnf['vnfDId'],'vnfInstanceId':vnf['vnfInstanceId']} for vnf in r['vnfs']]
            body = {'packageId':package_id,'vnfs':vnfs}
            count = 0
            start_time = time.time()
            r, s = makerequest(_type='post',url=f"{self.domain_orch_url}/terminate/vnf", body=body)
            if s == 201 and r['status'] == "TERMINATING":
                while True:
                    r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/terminate/vnf",params={'packageId':package_id})
                    if s==200 and r['status']=='TERMINATED':
                        end_time = time.time() - start_time
                        terminated = True
                        break
                    if count == len(vnfs) * 4:
                        raise typer.Exit(code=1)
                    count +=1
                    time.sleep(0.5)
            elif s == 500:
                raise typer.Exit(code=1)
            else:
                return r
        elif s == 200 and r['status'] in ['ONBOARDED','PARTIAL','INSTANTIATING','FAILED']:
            package_id = r['packageId']
            vim_id = r['vimId']
            terminated = True
        #OFFBOARD
        if terminated:
            body = {'packageId':package_id,'vimId':vim_id}
            count = 0
            r,s = makerequest(_type='post',url=f"{self.domain_orch_url}/offboard/vnf",body=body)
            if s == 201 and r['status'] == 'OFFBOARDING':
                while True:
                    r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/offboard/vnf",params={'packageId':package_id})
                    if s==200 and r['status']=='OFFBOARDED':
                        break
                    if count == 100: #it means timeout is 20 * 0.5 seconds
                        raise typer.Exit(code=1)
                    count +=1
                    time.sleep(0.5)


from owly_sdk.common.request_config import RequestConfig
from owly_sdk.vim.vim_manager import VimManager
from owly_sdk.ns.ns_manager import NsManager
from owly_sdk.mep.mep_manager import MepManager

class Owly:
    def __init__(self, domain_orch_url : str, request_config : RequestConfig = RequestConfig()):
        self.domain_orch_url = domain_orch_url
        self.request_config = request_config

        self.vim = VimManager(domain_orch_url, request_config)
        self.ns = NsManager(domain_orch_url, request_config)
        self.mep = MepManager(domain_orch_url, request_config)



from owly_sdk.common.utils import makerequest
import uuid
import time


class VnfManager:
    """
    The VnfManager class provides the functionality to manage VNFs such as
    onboarding, instantiation, termination, offboarding, listing, and getting logs.
    """
    def __init__(self, domain_orch_url, request_config):
        """
        Initialize the VnfManager class with the domain_orch_url and request_config.
        
        :param domain_orch_url: The URL of the Orchestrator.
        :param request_config: The RequestConfig object to be used for making requests.
        """
        self.domain_orch_url = domain_orch_url
        self.request_config = request_config
        

    def terminate(self, service, vnf):
        """
        Terminate a VNF of a network service
        
        :param service: network service configuration
        :param vnf: name of the VNF to be terminated
        
        :return: Termination status
        """

        terminated = False
        _r,_s = makerequest(_type='get',url=f"{self.domain_orch_url}/list/vnf",params={'name':service['name']})
        if _s == 200 and _r['status'] == 'NOTFOUND':
            raise Exception(f"Service {service['name']} is not found: {_r}")
        elif _s == 200 and _r['status'] in ['INSTANTIATED','TERMINATING']:
            package_id = _r['packageId']
            vim_id = _r['vimId']
            if vnf is not None:
                vnfs = [{'vnfName':v['vnfName'],'vnfDId':v['vnfDId'],'vnfInstanceId':v['vnfInstanceId']} for v in _r['vnfs'] if vnf in v['vnfName'] ]
                _vnfs = [v for v in service['vnfs'] if v['vnfName'] == vnf]
                service.update({'packageId':package_id,'vnfs':_vnfs})
            else:
                vnfs = [{'vnfName':v['vnfName'],'vnfDId':v['vnfDId'],'vnfInstanceId':v['vnfInstanceId']} for v in _r['vnfs']]
            if len(vnfs) == 0 or vnfs is None:
                raise Exception(f"VNF {vnf} is not instantiated")
            body = {'packageId':package_id,'vnfs':vnfs}
            count = 0
            start_time = time.time()
            r,s = makerequest(_type='post',url=f"{self.domain_orch_url}/terminate/vnf",body=body)
            if s == 201 and r['status'] == "TERMINATING":
                while True:
                    if vnf is not None:
                        params = {'packageId':package_id,'vnfInstanceId':vnfs[0]['vnfInstanceId']}
                    else:
                        params = {'packageId':package_id}
                    r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/terminate/vnf",params=params)
                    if s==200 and r['status'] == "TERMINATED":
                        end_time = time.time() - start_time
                        terminated = True
                        break
                    if count == len(vnfs) * 20:
                        raise Exception(f"Time out in terminating the service {service['name']}")
                    count +=1
                    time.sleep(0.5)
            else:
                raise Exception(f"Server replied with {s}, response {r}")

        return terminated


    def termination_status(self, packageId : str, vnfInstanceId : str = None):
        '''
        get the status of the termination process
        
        :param packageId: package id of the network service
        :param vnfInstanceId: instance id of the network function
        
        :return: status of the termination process
        '''
        r, s = makerequest(
            _type = 'get',
            url = f"{self.domain_orch_url}/terminate/vnf",
            params = {
                'packageId':packageId,
                'vnfInstanceId':vnfInstanceId
            }
        )
        
        return r


    def offboard(self, service : dict, vnf : str = None):
        """
        Offboard a VNF of a network service
        
        :param service: network service configuration
        :param vnf: name of the VNF to be offboarded
        
        :return: None
        """
        offboarded = False
        package_id = service['packageId']
        vim_id = service['vimId']
        body = {'packageId':package_id,'vimId':vim_id}
        if vnf is not None:
            vnfDIds = [v['vnfDId'] for v in service['vnfs'] if vnf in v['vnfName']]
            body.update({'vnfDIds':vnfDIds})
        else:
            return None
        count = 0
        start_time = time.time()
        r,s = makerequest(_type='post',url=f"{self.domain_orch_url}/offboard/vnf",body=body)
        if s == 201 and r['status'] == 'OFFBOARDING':
            while True:
                r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/list/vnf",params={'name':service['name'],'vnfName':vnf,'latest':'no'})
                if s==200 and (r == [] or r['status'] == 'NOTFOUND'):
                    end_time = time.time() - start_time
                    offboarded=True
                    break
                if count == 100: #it means timeout is 20 * 0.5 seconds
                    raise Exception(f"Time out in offboarding the vnf {vnf} for service {service['name']} ")
                count +=1
                time.sleep(0.5)
        return offboarded



    def offboarding_status(self, packageId : str):
        '''
        get the status of the offboarding process
        
        :param packageId: package id of the network service
        
        :return: status of the offboarding process
        '''
        r, s = makerequest(_type='get',url=f"{self.domain_orch_url}/offboard/vnf",params={'packageId':packageId})
        return r


    def onboard(self, service : dict, packageId : str):
        '''
        onboard network function to a network service

        :param service: network service configuration
        :param packageId: package id of the service
        
        return: network service configuration
        '''
        service.update({'packageId':packageId})
        [ vnf.update({'vnfDId': str(uuid.uuid4())}) for vnf in service['vnfs'] ]
        service.update({'vimId':[service['vimId']]})
        r, s = makerequest(_type='post',url= f"{self.domain_orch_url}/onboard/vnf",body=service)
        if s != 201:
            raise Exception(f"Encounted an error while posting the request {r}")
        r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/onboard/vnf",params={'packageId':service['packageId']})
        if not (s == 200 and r['status'] == 'ONBOARDED'):
            raise Exception(f"Server replied with status code {s}, response {r}")
        return service


    def onboarding_status(self, packageId : str):
        '''
        get the status of the onboarding process

        :param packageId: package id of the network service
        
        :return: status of the onboarding process
        '''
        r, s = makerequest(_type='get',url=f"{self.domain_orch_url}/onboard/vnf",params={'packageId':packageId})
        return r


    def instantiate(self, name : str, body : dict):
        '''
        instantiate network functions of a network service

        '''
        count = 0
        start_time = time.time()
        r, s = makerequest(_type='post',url=f"{self.domain_orch_url}/instantiate/vnf",body=body)
        print(r, s)
        if s !=201:
            print(f"Encounted an error, details: {r}")
            return None
        print(f"Made a request to instantiate {name} service")
        print(f"Instantiating {name} service")
        while True:
            r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/instantiate/vnf",params={'packageId':body['packageId']})
            if s == 200 and r['status'] != 'INSTANTIATED':
                print(f"Service in state {r['status']}\n{r['reason']}")
            elif s ==200 and r['status'] == 'INSTANTIATED':
                end_time = time.time() - start_time
                print(f"Service {r['status']} in {end_time} seconds")
                break
            else:
                print(f"Encounted an error at the server side, details: {r}")
                break
            if count == len(body['vnfs']) * 10: #it means timeout is 100*len(vnfs) * 0.5 seconds
                break
            count +=1
            time.sleep(0.5)
        if s == 500:
            print(f"Server replied with 500, response {r}")
            raise Exception(f"Server replied with 500, response {r}")
        elif count == len(body['vnfs']) * 100: #it means timeout is 100*len(vnfs) * 0.5 seconds
            print(f"Timeout in instantiating the service {name}")
            print(f"Last response from the server {r}")
            raise Exception(f"Timeout in instantiating the service {name}")
        elif s not in [500,200,201]:
            print(f"Encounted an error at the server side, details: {r['reason']}")
            
            
    def instantiation_status(self, packageId : str):
        '''
        get the status of the network service
        
        :param packageId: package id of the network service
        
        :return: status of the network service
        '''
        r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/instantiate/vnf",params={'packageId':packageId})
        return r


    def get_list(self, name : str, vnfName : str = None, latest : str = 'yes'):
        '''
        list all the network services
        
        :param name: name of the network service
        :param vnfName: name of the network function
        :param latest: get the latest network service
        
        :return: list of network services
        '''
        r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/list/vnf", params={'name':name, 'vnfName':vnfName, 'latest':latest})
        return r


    def logs(
        self,
        vim: str,
        name: str,
        vnf: str,
        containername: str = None,
        lines: int = None
    ):
        """
        Logs of a VNF of a network service NAME VNF_NAME
        
        :param vim: The name or ID of the VIM
        :param name: The name of the network service
        :param vnf: The name of the VNF
        :param containername: The name of the container
        :param lines: The number of lines to output
        
        :return: The logs of the VNF or an error message
        """
        params = {'name':name,'vnfName':vnf, 'vimId': vim, 'containerName':containername,'lines':lines}
        r, s = makerequest(_type='get',url=f"{self.domain_orch_url}/logs/vnf",params=params)

        if s in [200, 404, 400]:
            return r
        else:
            return "Encounted an error in fetching the list of onboarded services"
        

    def metrics(self, vim : str, name : str, vnf : str = None):
        """
        get the metrics of the network service
        
        :param vim: ID or name of the VIM
        :param name: name of the network service
        :param vnf: name of the network function
        
        :return: metrics of the network service
        """
        r, s = makerequest(_type='get',url=f"{self.domain_orch_url}/metrics/vnf",params={'name':name, 'vnfName':vnf, 'vimId':vim})
        
        if s in [200, 404, 400]:
            return r
        else:
            return "Encounted an error in fetching the list of onboarded services"


from owly_sdk.common.utils import makerequest
from pathlib import Path
import yaml
import json
import time
import typer
import uuid
import base64


class MecnsManager:
    """
    The MecnsManager class provides the functionality to manage MEC Application Services such as
    creating, listing, describing, and deleting MEC Application Services.
    """
    def __init__(self, domain_orch_url, request_config):
        self.domain_orch_url = domain_orch_url
        self.request_config = request_config
        

    def format_mecns(mecns):
        """
        Format the mecns response.
        
        :param mecns: The list of mecns
        
        :return: The list of formatted mecns
        """
        formatted_mecns = []

        keys = ['name', 'status', 'apps']
        app_keys = ['appName', 'status']
        for mecsn in mecns:
            formatted_apps = []
            
            for app in mecsn['apps']:
                formatted_app = {key: app[key] for key in app_keys}
                
                if app['status'] == 'INSTANTIATED':
                    formatted_app['interfaces'] = [
                        f"{networkDetail['ipaddress']}({networkDetail['interface']})"
                        for networkDetail in app['networkDetail']
                    ]
                formatted_apps.append(formatted_app)
                
            
            mecsn['apps'] = formatted_apps   
            formatted_mecsn = {key: mecsn[key] for key in keys}
            formatted_mecns.append(formatted_mecsn)
            
        return formatted_mecns


    def list_mecns(self, output: str = None):
        """
        List of mecsns registered in the database
        
        :param output: Output format (json/yaml)
        
        :return: List of mecsns
        """
        r, _ = makerequest(_type='get',url=f"{self.domain_orch_url}/list/app",params={'name':'all','latest':'yes'})
        
        formatted_mecsns = self.format_mecns(r)

        if output == 'yaml':
            return yaml.dump(formatted_mecsns)
        if output == 'json':
            json.dumps(formatted_mecsns)
        return formatted_mecsns

    def mecappservicecreate(self, service, file):
        '''
        onboard network functions of a mec application service
        
        :param service: mec application service configuration
        :param file: path to the yaml file containing the mec application service configuration
        
        :return: The mec application service configuration
        '''
        if 'packageId' not in service.keys():
            service.update({'packageId':str(uuid.uuid4())})
        for app in service['apps']:
            for container in app['osContainerDesc']:
                if 'bootData' in container.keys():
                    for data in container['bootData']:
                        data.update({'data': (base64.b64encode((open(Path(file.parent) / data['file']).read()).encode('ascii'))).decode('ascii')})
        [ app.update({'appDId': str(uuid.uuid4())}) for app in service['apps'] ]
        if not isinstance(service['vimId'],list):
            service.update({'vimId':[service['vimId']]})
        if 'mepId' in service.keys() and not isinstance(service['mepId'],list):
            service.update({'mepId':[service['mepId']]})
        r, s = makerequest(_type='post',url= f"{self.domain_orch_url}/create/mecs",body=service)
        if s!=201:
            raise typer.Exit(code=1)
        return service


    def create_mecns(self, file: Path):
        """
        Create a mec application service.
        Provide the path of the yaml file which 
        contains the mec application service defination, its apps.

        :param file: The path to the yaml file containing the mec application service defination.
        
        :return: Success message or error message.
        """
        try:
            with open(file) as f:
                services = yaml.safe_load_all(f)
                for service in services:
                    service = self.mecappservicecreate(service=service,file=file)        
        except Exception as e:
            return e


    def describe_mecns(self, name: str, output: str = 'json'):
        """
        Describe a MEC Application NAME

        :param name: The name of the MEC Application
        :param output: Output format (json/yaml). Default is 'json'
        
        :return: The MEC Application description
        """
        r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/list/app",params={'name':name})
        if s==200:
            if output == 'yaml':
                return yaml.dump(r)
            return json.dumps(r)
        else:
            return "Encounted an error in fetching the list of onboarded services"


    def top_mecns(self, vim: str, name: str, app: str = None, output: str = 'json'):
        """
        Performs top on MEC Application VIM_NAME NAME.
        Provides information related to computational 
        resources consumed by the MEC Application 

        :param vim: The name of the VIM
        :param name: The name of the MEC Application
        :param app: The name of the MEC Application Service
        :param output: Output format (json/yaml). Default is 'json'
        
        :return: None
        """
        r,s = makerequest(
            _type='get',
            url=f"{self.domain_orch_url}/metrics/app",
            params={'name':name,'appName':app,'vimId':vim}
        )
        if s in [200,404,400]:
            if output == 'yaml':
                return yaml.dump(r)
            return json.dumps(r)
        else:
            return "Encounted an error in fetching the list of onboarded services"


    def app_logs(
        self,
        vim: str,
        name: str,
        app: str,
        containername: str = None,
        lines: int = None
    ):
        """
        Logs of a MEC Application NAME app_NAME
        
        :param vim: The VIM ID
        :param name: The name of the MEC Application
        :param app: The name of the MEC Application Service
        :param containername: The name of the container
        :param lines: The number of lines to output
        
        :return: The logs of the MEC Application Service or an error message
        """
        params = {'name':name,'appName':app, 'vimId': vim, 'containerName':containername,'lines':lines}
        r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/logs/app",params=params)
        if s in [200,404,400]:
            return r
        else:
            return "Encounted an error in fetching the list of onboarded services"

        
    def delete_mecns(self, name: str):
        """
        Delete the mec application service
        
        :param name: The name of the mec application service
        
        :return: None
        """
        terminated = False
        r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/list/app",params={'name':name,'latest':'no'})
        if s == 200 and r['status'] == 'NOTFOUND':
            raise typer.Exit(code=1)
        elif s == 200 and r['status'] in ['INSTANTIATED','TERMINATING']:
            package_id = r['packageId']
            vim_id = r['vimId']
            apps = [{'appName':app['appName'],'appDId':app['appDId'],'appInstanceId':app['appInstanceId']} for app in r['apps']]
            body = {'packageId':package_id,'apps':apps}
            count = 0
            start_time = time.time()
            r,s = makerequest(_type='post',url=f"{self.domain_orch_url}/terminate/app",body=body)
            if s == 201 and r['status'] == "TERMINATING":
                while True:
                    r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/terminate/app",params={'packageId':package_id})
                    if s==200 and r['status']=='TERMINATED':
                        end_time = time.time() - start_time
                        terminated = True
                        break
                    if count == len(apps) * 4:
                        raise typer.Exit(code=1)
                    count +=1
                    time.sleep(0.5)
            elif s == 500:
                raise typer.Exit(code=1)
        elif s == 200 and r['status'] in ['ONBOARDED','PARTIAL','INSTANTIATING','FAILED']:
            package_id = r['packageId']
            vim_id = r['vimId']
            terminated = True
        #OFFBOARD
        if terminated:
            body = {'packageId':package_id,'vimId':vim_id}
            count = 0
            start_time = time.time()
            r,s = makerequest(_type='post',url=f"{self.domain_orch_url}/offboard/app",body=body)
            if s == 201 and r['status'] == 'OFFBOARDING':
                while True:
                    r,s = makerequest(_type='get',url=f"{self.domain_orch_url}/offboard/app",params={'packageId':package_id})
                    if s==200 and r['status']=='OFFBOARDED':
                        end_time = time.time() - start_time
                        break
                    if count == 100: #it means timeout is 20 * 0.5 seconds
                        raise typer.Exit(code=1)
                    count +=1
                    time.sleep(0.5)


from owly_sdk.common.utils import makerequest
from pathlib import Path
import time
import yaml


class VimManager:
    """
    The VimManager class provides the functionality to manage VIMs such as listing, adding, describing, and removing VIMs.
    """
    def __init__(self, domain_orch_url, request_config):
        """
        Initialize the VimManager class with the domain_orch_url and request_config.
        
        :param domain_orch_url: The URL of the Orchestrator.
        :param request_config: The RequestConfig object to be used for making requests.
        """
        self.domain_orch_url = domain_orch_url
        self.request_config = request_config

    def _format_vim(self, vims: list, wide: bool = False):
        """
        Format the VIMs to display only the required information.

        :param vims: The list of VIMs to be formatted.
        :param wide: Whether to display more information about the VIMs. Default is False.
        
        :return: The formatted VIMs.
        """
        formatted_vims = []
        for vim in vims:
            keys = ['name', 'status', 'url']
            wide_keys = ['vimType']
            
            formatted_response = {k: vim[k] for k in keys}

            if wide:
                for k in wide_keys:
                    formatted_response[k] = vim[k]
            
            formatted_vims.append(formatted_response)
            
        return formatted_vims

    def list_vims(self, vim_id: str = 'all', wide: bool = False):
        """
        Fetch the list of VIMs registered in the database.

        :param vim_id: ID of the VIM to be listed. If not provided, all VIMs will be listed.
        :param wide: Whether to display more information about the VIMs. Default is False.
        
        :return: List of VIMs registered in the database or a specific VIM if vim_id is provided.
        """
        response, _ = makerequest(
            _type='get',
            url=f"{self.domain_orch_url}/resource/vim",
            params={'vimId': vim_id},
            config=self.request_config
        )
        
        if isinstance(response, dict):
            if 'status' in response and response['status'] == "NOTFOUND":
                return response
            response = [response]
        
        formatted_vims = self._format_vim(response, wide)
        
        if vim_id == 'all':
            return formatted_vims
        
        return formatted_vims[0]

    def add_vim(self, file: Path):
        """
        Add a VIM to the list of registered VIMs.
        Provide the path of the yaml file which
        contains the VIM(s) definition.
        You can add multiple VIMs at the same time if mentioned in the same file.

        :param file: The path of the YAML file containing the VIM(s) definition.

        :return: A string message indicating the success or failure of the registration process.
        """
        start_time = time.time()
        with open(file) as f:
            vims = yaml.safe_load_all(f)
            for vim in vims:
                if 'kind' not in vim.keys() or vim['kind'] != 'vim':
                    return 'Not a valid VIM template, kind is not vim'
                r, s = makerequest(
                    _type='post',
                    url=f"{self.domain_orch_url}/resource/vim",
                    body=vim,
                    config=self.request_config
                )
                if s == 201 and r['status'] == 'REGISTERED':
                    end_time = time.time() - start_time
                    return f"VIM {vim['name']} successfully registered in {end_time} seconds"
                else:
                    return f"VIM {vim['name']} cannot be registered: {r['reason']}"

    def describe_vim(self, name: str):
        """
        Describe the current state of the VIM NAME.
        
        :param name: The name of the VIM to be described.
        
        :return: The response from the API.
        """
        response, _ = makerequest(
            _type='get',
            url=self.domain_orch_url + '/resource/vim',
            params={'vimId': name},
            config=self.request_config
        )
        
        return response

    def remove_vim(self, name: str, forcefully: bool = False):
        """
        Remove the VIM NAME from the list of registered VIMs.
        Normally if there are workloads running on the VIM, 
        it is not possible to remove it. 
        
        :param name: The name of the VIM to be removed.
        :param forcefully: If True, the VIM will be removed forcefully. Default is False.
        
        :return: A string message indicating the success or failure of the removal process.
        """
        response, status_code = makerequest(
            _type='delete',
            url=self.domain_orch_url + '/resource/vim',
            params={'vimId': name, 'forcefully': forcefully, 'wipe': False},
            config=self.request_config
        )

        if status_code == 204:
            return f"VIM {name} successfully removed"      
        else:
            if 'instances' in response.keys():
                return f"VIM {name} could not be removed because {response['reason']} instances {response['instances']}"
            else:
                return f"VIM {name} could not be removed because {response['reason']}"



"""

export PYTHONPATH=/home/foreur/Desktop/thomas/owly-sdk

"""

import unittest
import src.vim.vim_manager as vim

VIM_FILE = "/home/foreur/Desktop/thomas/blueprints/mec-gps-service/vim-edge.yaml"

class TestMECNSManagement(unittest.TestCase):
    
    def setUp(self):
        """Setup initial conditions."""
        self.vim_name = "edge"
        
        # Check if the VIM exists
        self.existing_vims = vim.list_vims()
        self.assertIsInstance(self.existing_vims, list, "The return value should be a list")
        # Delete the VIM if it exists
        if any(vim["name"] == self.vim_name for vim in self.existing_vims):
            vim.remove_vim(self.vim_name)
            
            
    def tearDown(self):
        """Clean up after the test."""
        # Delete all VIMs
        for v in vim.list_vims():
            vim.remove_vim(v["name"])
            

    def test_list_vims_returns_list(self):
        """Test that list_vims returns a list."""
        vims = vim.list_vims()
        self.assertIsInstance(vims, list, "The return value should be a list")

    def test_list_vims_returns_list_of_dicts(self):
        """Test that list_vims returns a list of dictionaries."""
        vims = vim.list_vims()
        for v in vims:
            self.assertIsInstance(v, dict, "Each element in the list should be a dictionary")
            
    def test_add_vim(self):
        """Test that add_vim adds a VIM."""
        res = vim.add_vim(VIM_FILE)
        
        vims = vim.list_vims()
        # Check if the VIM was added
        self.vim_name = "edge"
        self.assertTrue(any(v["name"] == self.vim_name for v in vims), "The VIM should have been added")
        
    def test_remove_vim(self):
        """Test that remove_vim removes a VIM."""
        res = vim.add_vim(VIM_FILE)
        
        vims = vim.list_vims()
        # Check if the VIM was added
        self.vim_name = "edge"
        self.assertTrue(any(v["name"] == self.vim_name for v in vims), "The VIM should have been added")
        
        res = vim.remove_vim(self.vim_name)
        
        vims = vim.list_vims()
        # Check if the VIM was removed
        self.assertFalse(any(v["name"] == self.vim_name for v in vims), "The VIM should have been removed")

if __name__ == "__main__":
    unittest.main()


from owly_sdk.common.utils import makerequest, format_response
from owly_sdk.common.constants import DOMAIN_ORCH_BASE_URL
from pathlib import Path
import yaml
import time


class MepManager:
    """
    The MepManager class provides the functionality to manage MEPs such as listing, adding, describing, and removing MEPs.
    """
    def __init__(self, domain_orch_url, request_config):
        """
        Initialize the MepManager class with the domain_orch_url and request_config.
        
        :param domain_orch_url: The URL of the Orchestrator.
        :param request_config: The RequestConfig object to be used for making requests.
        """
        self.mep_url = f"{domain_orch_url}/resource/mep"
        self.request_config = request_config


    def add_mep(self, file: Path):
        """
        Add a mep to the list of registered meps.
        Provide the path of the yaml file which
        contains the mep(s) defination.
        You can add multiple meps at the same time. If mentioned in the same file.
        
        :param file: The path to the yaml file containing the mep(s) defination.
        
        :return: Success message or error message.
        """
        start_time = time.time()
        with open(file) as f:
            meps = yaml.safe_load_all(f)
            for mep in meps:
                r, s = makerequest(_type='post',url=self.mep_url,body=mep)
                if s == 201 and r['status'] == 'REGISTERED':
                    end_time = time.time() - start_time
                    return f"mep {mep['name']} succesfully registered in {end_time} seconds"
                else:
                    if 'reason' in r.keys():
                        return f"mep {mep['name']} can not be registered reason {r['reason']}"
                    return r


    def list_meps(self, output: str = None):
        """
        List of meps registered in the database
        
        :param output: Output format (json/yaml).
        
        :return: List of meps
        """
        response, _ = makerequest(_type='get', url=self.mep_url, params={'mepId':'all'})
        
        formatted_meps = []

        keys = ['name', 'status']
        for mep in response:
            formatted_mep = {key: mep[key] for key in keys}
            
            formatted_meps.append(formatted_mep)
            
        return format_response(formatted_meps, output)


    def describe_mep(self, name: str, output: str = None):
        """
        Describe the current state of the mep NAME. 

        :param name: The name of the mep to describe.
        :param output: Format of the output (json/yaml).
        
        :return: The response from the API or an error message.
        """
        response, _ = makerequest(_type='get', url=self.mep_url, params={'mepId':name})

        return format_response(response, output)


    def remove_mep(self, name: str):
        """
        Remove the mep NAME from the list of registered meps.
        Normally if there are workloads running on the mep 
        it is not possible to remove it.
        
        :param name: The name of the mep to remove.
        
        :return: Response from the API.
        """
        response, _ = makerequest(_type='delete', url=self.mep_url, params={'mepId':name})

        return response